@@ -9,7 +9,14 @@
 # Copyright (c) 2021 transentis labs GmbH
 # MIT License
 
-from flask import Flask, redirect, url_for, request, make_response, jsonify
+import sys
+version = sys.version_info
+if(version[0] < 3 or (version[0] == 3 and version[1] < 9)):
+    print("BPTK Server requires Python 3.9 or later. Please update Python to use the BPTK Server! Exitting now.")
+    sys.exit()
+
+
+from flask import Flask, redirect, url_for, request, make_response, jsonify, Response
 from BPTK_Py.bptk import bptk
 import pandas as pd
 import json
@@ -19,6 +26,7 @@ import datetime
 from json import JSONEncoder
 import jsonpickle
 import copy
+import threading
 from BPTK_Py.externalstateadapter import InstanceState, ExternalStateAdapter
 
 class InstanceManager:
@@ -43,7 +51,7 @@ class InstanceManager:
     def _get_instance_state(self, instance_uuid):
         instance = self._instances[instance_uuid]
         session_state = copy.deepcopy(instance['instance'].session_state)
-        
+        session_state["lock"] = False
         return InstanceState(session_state, instance_uuid, instance["time"], instance["timeout"], session_state["step"])
             
     def get_instance_states(self):
@@ -91,13 +99,20 @@ class InstanceManager:
             }
 
         metrics["instanceCount"] = len(self._instances)
+        metrics["threadCount"] = threading.active_count()
 
         return metrics
 
         
     def _get_prometheus_instance_metrics(self):
         self._timeout_instances()
-        return "# HELP bptk_instance_count The number of instances in the beergame server\n# TYPE bptk_instance_count gauge\nbptk_instance_count " + str(len(self._instances))
+        metrics =  "# HELP bptk_instance_count The number of instances in the bptk server\n# TYPE bptk_instance_count gauge\nbptk_instance_count " + str(len(self._instances)) + "\n"
+        metrics += "# HELP bptk_thread_count The number of threads in the bptk server\n# TYPE bptk_thread_count gauge\nbptk_thread_count " + str(threading.active_count()) + "\n"
+        return metrics
+
+    def _delete_instance(self, instance_id):
+        if instance_id in self._instances:
+            del self._instances[instance_id]
 
     def create_instance(self,**timeout):
         """
@@ -200,6 +215,8 @@ class BptkServer(Flask):
         self.route("/agents", methods=['POST', 'PUT'], strict_slashes=False)(self._agents_resource)
         self.route("/start-instance", methods=['POST'], strict_slashes=False)(self._start_instance_resource)
         self.route("/<instance_uuid>/run-step", methods=['POST'], strict_slashes=False)(self._run_step_resource)
+        self.route("/<instance_uuid>/run-steps", methods=['POST'], strict_slashes=False)(self._run_steps_resource)
+        self.route("/<instance_uuid>/stream-steps", methods=['POST'], strict_slashes=False)(self._stream_steps_resource)
         self.route("/<instance_uuid>/begin-session", methods=['POST'], strict_slashes=False)(self._begin_session_resource)
         self.route("/<instance_uuid>/end-session", methods=['POST'], strict_slashes=False)(self._end_session_resource)
         self.route("/<instance_uuid>/session-results", methods=['GET'], strict_slashes=False)(self._session_results_resource)
@@ -209,6 +226,17 @@ class BptkServer(Flask):
         self.route("/full-metrics", methods=['GET'], strict_slashes=False)(self._full_metrics)
         self.route("/save-state", methods=['GET'], strict_slashes=False)(self._save_state_resource)
         self.route("/load-state", methods=['GET'], strict_slashes=False)(self._load_state_resource)
+        self.route("/<instance_uuid>/stop-instance", methods=['GET'], strict_slashes=False)(self._stop_instance)
+
+    def _stop_instance(self, instance_uuid):
+        self._instance_manager._delete_instance(instance_uuid)
+        if self._external_state_adapter != None:
+            self._external_state_adapter.delete_instance(instance_uuid)
+
+        resp = make_response("Instance deleted.", 200)
+        resp.headers['Content-Type']='application/json'
+        resp.headers['Access-Control-Allow-Origin']='*'
+        return resp
 
     def _save_state_resource(self):
         """
@@ -530,21 +558,44 @@ class BptkServer(Flask):
             resp.headers['Content-Type']='application/json'
             resp.headers['Access-Control-Allow-Origin']='*'
             return resp
+        equations = []
+        agents = []
+        agent_states=[]
+        agent_properties=[]
+        agent_property_types=[]
+        individual_agent_properties=[]
 
-        try:
-            equations = content["equations"]
-        except KeyError:
-            resp = make_response('{"error": "expecting equations to be set"}', 500)
+        if(not "agents" in content.keys() and not "equations" in content.keys()):
+            resp = make_response('{"error": "expecting either equations or agents to be set"}', 500)
             resp.headers['Content-Type']='application/json'
             resp.headers['Access-Control-Allow-Origin']='*'
             return resp
+        if("agents" in content.keys()):
+            agents = content["agents"]
+        if("equations" in content.keys()):
+            equations = content["equations"]
+        if("agent_states" in content.keys()):
+            agent_states = content["agent_states"]
+        if("agent_properties" in content.keys()):
+            agent_properties = content["agent_properties"]
+        if("agent_property_types" in content.keys()):
+            agent_property_types = content["agent_property_types"]
+        if("individual_agent_properties" in content.keys()):
+            individual_agent_properties = content["individual_agent_properties"]
+
+
 
         instance = self._instance_manager.get_instance(instance_uuid)
 
         instance.begin_session(
             scenario_managers=scenario_managers,
             scenarios=scenarios,
-            equations=equations
+            equations=equations,
+            agents=agents,
+            agent_states=agent_states,
+            agent_properties=agent_properties,
+            agent_property_types=agent_property_types,
+            individual_agent_properties=individual_agent_properties
         )
 
         resp = make_response('{"msg":"session started"}', 200)
@@ -608,10 +659,8 @@ class BptkServer(Flask):
         This endpoint advances the relevant scenarios by one timestep and returns the data for that timestep.
 
         Arguments:
-            settings: JSON
-                Dictionary structure with a key "settings" that contains the settings to apply for that step. These can be constants and points.
-            flatResults: boolean
-                When set to true, a flat version of the results is returned.
+            instance_uuid: string
+                The id of the instance to advance.
         """
         # Checking if the instance id is valid.
         if not self._ensure_instance_exists(instance_uuid):
@@ -621,6 +670,13 @@ class BptkServer(Flask):
             return resp
         
         instance = self._instance_manager.get_instance(instance_uuid)
+
+        if(instance.is_locked()):
+            resp = make_response('{"error": "instace is locked"}', 500)
+            resp.headers['Content-Type'] = 'application/json'
+            resp.headers['Access-Control-Allow-Origin'] = '*'
+            return resp
+
         if not request.is_json:
             result = instance.run_step()
         else:
@@ -634,7 +690,7 @@ class BptkServer(Flask):
                 return resp
 
         if result is not None:
-            resp = make_response(json.dumps(result), 200)
+            resp = make_response(jsonpickle.dumps(result), 200)
         else:
             resp = make_response('{"error": "no data was returned from run_step"}', 500)
 
@@ -645,6 +701,133 @@ class BptkServer(Flask):
         resp.headers['Access-Control-Allow-Origin']='*'
         return resp
 
+    def _run_steps_resource(self, instance_uuid):
+        """
+        This endpoint advances the relevant scenarios by one timestep and returns the data for that timestep.
+
+        Arguments:
+            instance_uuid: string
+                The id of the instance to advance.
+        """
+        # Checking if the instance id is valid.
+        if not self._ensure_instance_exists(instance_uuid):
+            resp = make_response('{"error": "expecting a valid instance id to be given"}', 500)
+            resp.headers['Content-Type'] = 'application/json'
+            resp.headers['Access-Control-Allow-Origin'] = '*'
+            return resp
+        
+        result = []
+        try:
+            instance = self._instance_manager.get_instance(instance_uuid)
+            if not request.is_json:
+                resp = make_response('{"error": "expecting a number of steps to be provided in the body as a json {numberSteps: int}"}', 500)
+                resp.headers['Content-Type'] = 'application/json'
+                resp.headers['Access-Control-Allow-Origin'] = '*'
+                return resp
+
+            if(instance.is_locked()):
+                resp = make_response('{"error": "instace is locked"}', 500)
+                resp.headers['Content-Type'] = 'application/json'
+                resp.headers['Access-Control-Allow-Origin'] = '*'
+                return resp
+            content = request.get_json()
+            if "numberSteps" in content:
+                if "settings" in content:
+                    instance.lock()
+                    for i in range(0,content["numberSteps"]):
+                        result.append(instance.run_step(settings=content["settings"], flat="flatResults" in content and content["flatResults"] == True))
+                    instance.unlock()
+                else:
+                    resp = make_response('{"error": "expecting settings to be set"}', 500)
+                    resp.headers['Content-Type'] = 'application/json'
+                    resp.headers['Access-Control-Allow-Origin'] = '*'
+                    return resp
+            else:
+                resp = make_response('{"error": "expecting a number of steps to be provided in the body as a json {"numberSteps": int}"}', 500)
+                resp.headers['Content-Type'] = 'application/json'
+                resp.headers['Access-Control-Allow-Origin'] = '*'
+                return resp
+        except:
+            instance.unlock()
+        if result is not None:
+            resp = make_response(jsonpickle.dumps(result), 200)
+        else:
+            resp = make_response('{"error": "no data was returned from run_step"}', 500)
+
+        if self._external_state_adapter != None:
+            self._external_state_adapter.save_instance(self._instance_manager._get_instance_state(instance_uuid))
+
+        resp.headers['Content-Type'] = 'application/json'
+        resp.headers['Access-Control-Allow-Origin']='*'
+        return resp
+
+
+
+    def _stream_steps_resource(self, instance_uuid):
+        """
+        This endpoint is used to stream a simulation.
+
+        Arguments:
+            
+            instance_uuid: string
+                The id of the instance to stream.
+        """
+        # Checking if the instance id is valid.
+        if not self._ensure_instance_exists(instance_uuid):
+            resp = make_response('{"error": "expecting a valid instance id to be given"}', 500)
+            resp.headers['Content-Type'] = 'application/json'
+            resp.headers['Access-Control-Allow-Origin'] = '*'
+            return resp
+        
+        instance = self._instance_manager.get_instance(instance_uuid)
+        is_json = request.is_json
+        if is_json:
+            content = request.get_json()
+            
+            if not "settings" in content:
+                resp = make_response('{"error": "expecting settings to be set"}', 500)
+                resp.headers['Content-Type'] = 'application/json'
+                resp.headers['Access-Control-Allow-Origin'] = '*'
+                return resp
+
+        if(instance.is_locked()):
+            resp = make_response('{"error": "instace is locked"}', 500)
+            resp.headers['Content-Type'] = 'application/json'
+            resp.headers['Access-Control-Allow-Origin'] = '*'
+            return resp
+
+        def streamer():
+            try:
+                instance.lock()
+                yield "["
+                first = True
+                while instance.progress() <= 1.0:
+                    if first:
+                        first = False
+                    else:
+                        yield ","
+                        
+                    if is_json:
+                        result = instance.run_step(settings=content["settings"], flat="flatResults" in content and content["flatResults"] == True)
+                    else:
+                        result = instance.run_step()
+                    if result is not None:
+                        yield jsonpickle.dumps(result)
+                    else:
+                        yield '{"error": "no data was returned from run_step"}'
+                yield "]"
+            except:
+                instance.unlock()
+            if self._external_state_adapter != None:
+                self._external_state_adapter.save_instance(self._instance_manager._get_instance_state(instance_uuid))
+
+        resp = Response(streamer())
+        resp.headers['Content-Type'] = 'application/json'
+        resp.headers['Access-Control-Allow-Origin'] = '*'
+        return resp
+
+
+
     def _keep_alive_resource(self,instance_uuid):
         """
         This endpoint sets the "last accessed time" of the instance to the current time to prevent the instance from timeing out.
